from mathutils import Vector
import struct
import bmesh
import bpy

from bpy_extras.io_utils import axis_conversion

def WriteNumber( Writer, Number, Length ) :
    Result = Number.to_bytes( Length, 'little' )
    
    for i in range( 0, Length ) :
        Writer.append( Result[ i ] )
            
    return None

def WriteFloat( Writer, Float ) :
    Result = bytearray( struct.pack( 'f', Float ) )
    
    for i in range( 0, 4 ) :
        Writer.append( Result[ i ] )
            
    return None

def WriteStrRaw( Writer, String ) :
    for i in range( 0, len( String ) ) :
        Writer.append( ord( String[ i ] ) )
    
    return None

def WriteStrLen( Writer, String ) :
    WriteNumber( Writer, len( String ), 4 )
    WriteStrRaw( Writer, String )    
    return None

def GetTriangulatedMesh( Object ) :
    try :
        Mesh = Object.to_mesh( Object.users_scene[ 0 ], True, 'PREVIEW', calc_tessface = False )        
    except :
        return None
   
    BiMesh = bmesh.new()
    BiMesh.from_mesh( Mesh )
    bmesh.ops.triangulate( BiMesh, faces = BiMesh.faces )
    BiMesh.to_mesh( Mesh )
    BiMesh.free()
    
    return Mesh 

def List_Contains( List, Value ) :
    try :
        Index = List.index( Value ) 
    except :
        Index = -1
    
    return Index != -1

# 0 - Nothing
# 1 - prev UV+1 - UV Zero fill
# 2 - prev UV+1 - UV Zero fill
# 4 - post UV+1 - UV Zero fill
# 8 - post UV+1 - UV Zero fill
def GetShaderFlags( ShaderID ) :
    if ShaderID == 'StadiumGrass' :
        return 0
    elif ShaderID == 'StadiumWater' :
        return 0
    elif ShaderID == 'StadiumDirt' :
        return 1
    elif ShaderID == 'StadiumRoad' :
        return 12
    elif ShaderID == 'StadiumRoadDetails' :
        return 12
    elif ShaderID == 'Dev5' :
        return 12
    elif ShaderID == 'Dev7' :
        return 0
    
    return 4

def GetSurfaceIndex( CollisionMaterial ) :
    CollisionMaterials = [
        'Concrete', 'Pavement',
        'Grass', 'Ice',
        'Metal', 'Sand',
        'Dirt', 'Turbo',
        'DirtRoad', 'Rubber',
        'SlidingRubber', 'Test',
        'Rock', 'Water',
        'Wood', 'Danger',
        'Asphalt', 'WetDirtRoad',
        'WetAsphalt', 'WetPavement',
        'WetGrass', 'Snow', 
        'ResonantMetal', 'GolfBall', 
        'GolfWall', 'GolfGround', 
        'TurboRed', 'Bumper', 
        'NotCollidable', 'FreeWheeling',
        'TurboRoulette'
    ]
    
    try :
        return CollisionMaterials.index( CollisionMaterial )
    except :
        return -1

#
# This function writes single CPlugTree
#
def DoWriteTree ( Writer, OriginObject, Object, MaterialInfo, InstancesNb ) :    
    ObjectMesh = GetTriangulatedMesh( Object )
    InstancesNb += 1
    
    WriteNumber( Writer, 0x0904F000, 4 )
    
    if MaterialInfo != None :
        WriteNumber( Writer, 0x0904F00D, 4 )
        
        if Object[ 'Index' ] == 0 :
            WriteNumber( Writer, 0x00000003, 4 )
        
        WriteNumber( Writer, 0x40000000, 4 )
        WriteStrLen( Writer, MaterialInfo[ 'ExtName' ] )
        
        WriteNumber( Writer, 0xFFFFFFFF, 4 )
    
    WriteNumber( Writer, 0x0904F016, 4 )
    WriteNumber( Writer, InstancesNb, 4 )   
    
    # 09 01E 000 -- Start, CPlugVisualIndexedTriangles        
    WriteNumber( Writer, 0x0901E000, 4 )    

    LayerUVs = []
    Vertices = []
    Loops = []
    
    if len( ObjectMesh.uv_layers ) is 0 :
        for Vertex in ObjectMesh.vertices :
            Vertices.append( Vertex )
        
        for Loop in ObjectMesh.loops :
            Loops.append( Loop.vertex_index )    
    else :
        DictonaryUVs = {}
        UniqueUVs = 0
        LayerUV = []
        
        for Face in ObjectMesh.polygons :
            
            for LoopIndex in Face.loop_indices :
                MapData = ObjectMesh.uv_layers.active.data[ LoopIndex ].uv
                
                Key = ObjectMesh.loops[ LoopIndex ].vertex_index, ( round( MapData.x, 4 ), round( MapData.y, 4 ) )
                Value = DictonaryUVs.get( Key )
                
                if Value is None :
                    Vertices.append( ObjectMesh.vertices[ ObjectMesh.loops[ LoopIndex ].vertex_index ] )
                    Value = DictonaryUVs[ Key ] = UniqueUVs
                    LayerUV.append( MapData )
                    UniqueUVs += 1
                
                Loops.append( Value )
        
        MaterialIntFlags = 0
        
        if MaterialInfo != None :
            MaterialIntFlags = MaterialInfo[ 'IntFlags' ]
        
        if MaterialIntFlags & 1 :
            LayerUVs.append( [ Vector( [ 0, 0 ] ) for x in range( len( LayerUV ) ) ] )
        
        if MaterialIntFlags & 2 :
            LayerUVs.append( [ Vector( [ 0, 0 ] ) for x in range( len( LayerUV ) ) ] )
        
        LayerUVs.append( LayerUV )
        
        if MaterialIntFlags & 4 :
            LayerUVs.append( [ Vector( [ 0, 0 ] ) for x in range( len( LayerUV ) ) ] )
            
        if MaterialIntFlags & 8 :
            LayerUVs.append( [ Vector( [ 0, 0 ] ) for x in range( len( LayerUV ) ) ] )
        
        del UniqueUVs, DictonaryUVs
    
    # 09 006 00E -- Start  
    WriteNumber( Writer, 0x0900600E, 4 )
    WriteNumber( Writer, 0x00000038, 4 )                                # VisualFlags
    WriteNumber( Writer, len( LayerUVs ), 4 )                           # UVsCount
    WriteNumber( Writer, len( Vertices ), 4 )                           # VertexNb
    WriteNumber( Writer, 0x00000000, 4 )                                # Unknown

    for MapLayer in LayerUVs :
        WriteNumber( Writer, 0x00000000, 4 )
        
        for LayerData in MapLayer :
            WriteFloat( Writer, LayerData.x )
            WriteFloat( Writer, LayerData.y )

    WriteFloat( Writer, Object.dimensions.x / 2 )                       # GlobalHalfDiagX?
    WriteFloat( Writer, Object.dimensions.z / 2 )                       # GlobalHalfDiagY?
    WriteFloat( Writer, Object.dimensions.y / 2 )                       # GlobalHalfDiagZ?
    WriteFloat( Writer, Object.dimensions.x / 2 )                       # LocalHalfDiagX?
    WriteFloat( Writer, Object.dimensions.z / 2 )                       # LocalHalfDiagY?
    WriteFloat( Writer, Object.dimensions.y / 2 )                       # LocalHalfDiagZ?                    
    WriteNumber( Writer, 0x00000000, 4 )                                # Unknown
    
    # 09 02C 004 -- Start, VerticesData
    WriteNumber( Writer, 0x0902C004, 4 )
    
    for VertexData in Vertices :
        #VertexPos = Object.matrix_world * VertexData.co
        #VertexPos.y -= OriginObject.dimensions.y
        VertexPos = Object.matrix_local * VertexData.co
        
        WriteFloat( Writer, VertexPos.x )                               # Vertex X
        WriteFloat( Writer, VertexPos.z )                               # Vertex Y
        WriteFloat( Writer, -VertexPos.y )                              # Vertex Z
        WriteFloat( Writer, VertexData.normal.x )                       # Normal X
        WriteFloat( Writer, VertexData.normal.z )                       # Normal Y
        WriteFloat( Writer, -VertexData.normal.y )                      # Normal Z
        WriteFloat( Writer, Object.color[ 0 ] )                         # Color R
        WriteFloat( Writer, Object.color[ 1 ] )                         # Color G
        WriteFloat( Writer, Object.color[ 2 ] )                         # Color B
        WriteFloat( Writer, Object.color[ 3 ] )                         # Color A

    WriteNumber( Writer, 0x00000000, 4 )                                # Unknown
    WriteNumber( Writer, 0x00000000, 4 )                                # Unknown
    
    # 09 06A 001 -- Start
    WriteNumber( Writer, 0x0906A001, 4 )    
    WriteNumber( Writer, 0x00000001, 4 )                                # Unknown
    
    # 09 057 000 -- Start
    WriteNumber( Writer, 0x09057000, 4 )
    WriteNumber( Writer, 0x00000002, 4 )                                # Unknown
    WriteNumber( Writer, len( Loops ), 4 )                              # List size
    
    for Loop in Loops  :
        WriteNumber( Writer, Loop, 2 )
    
    WriteNumber( Writer, 0xFACADE01, 4 )
    # 09 057 000 -- End
    # 09 06A 001 -- End
    # 09 02C 004 -- End
    # 09 006 00E -- End
    
    WriteNumber( Writer, 0xFACADE01, 4 )
    # 09 01E 000 -- End
    
    WriteNumber( Writer, 0xFFFFFFFF, 4 )
    WriteNumber( Writer, 0xFFFFFFFF, 4 )
    WriteNumber( Writer, 0xFFFFFFFF, 4 )
    WriteNumber( Writer, 0xFACADE01, 4 )
    
    bpy.data.meshes.remove( ObjectMesh )
    del Vertices, LayerUVs, Loops
    return InstancesNb

#
# This function writes a couple of CPlugTrees basing on Objects argument
#
def DoWriteObject ( Writer, OriginObject, Objects, MaterialInfos, InstancesNb ) :
    WriteNumber( Writer, 0x0904F000, 4 )
    
    WriteNumber( Writer, 0x0904F006, 4 )
    WriteNumber( Writer, 0x0000000A, 4 )        # Unknown
    WriteNumber( Writer, len( Objects ), 4 )    # Childs

    for Object in Objects :
        MaterialInfo = None
        
        if len( Object.data.materials ) != 0 :
            MaterialName = Object.data.materials[ 0 ].name
            
            for MatInfo in MaterialInfos :
                if MatInfo[ 'IntName' ] != MaterialName :
                    continue
                
                MaterialInfo = MatInfo
                break
        
        InstancesNb += 1
        WriteNumber( Writer, InstancesNb, 4 )        
        InstancesNb = DoWriteTree( Writer, OriginObject, Object, MaterialInfo, InstancesNb )

    WriteNumber( Writer, 0xFACADE01, 4 )
    return InstancesNb

def DoWriteBlock ( Context, FilePath, SpawnX, SpawnY, SpawnZ, BlockIdentifier, BlockAuthor ) :
    if not bpy.ops.object.mode_set.poll() :
        return { 'CANCELLED' }
    
    if len( BlockIdentifier ) == 0 or len( BlockAuthor ) == 0 :
        return { 'CANCELLED' }
    
    bpy.ops.ed.undo_push( message = 'Export GBX' )
    bpy.ops.ed.undo_push( message = 'Export GBX' )
    
    bpy.ops.object.mode_set( mode = 'OBJECT' )
    bpy.ops.object.select_all( action= 'DESELECT' )

    Objects = Context.visible_objects
    Scene = Context.scene

    MaterialInfos = []
    MaterialsNb = 0
    
    for Object in Objects :
        Materials = Object.data.materials
        ObjectMaterials = []
        
        for Index in range( 0, len( Materials ) ) :
            Material = Materials[ Index ]            
            MaterialInfo = { 
                'ExtFlags' : 0,
                'IntFlags': 0,
            }
            
            if 'TextureGame' in Material :
                MaterialInfo[ 'IntFlags' ] = GetShaderFlags( Material[ 'TextureGame' ] )
                MaterialInfo[ 'TextureID' ] = Material[ 'TextureGame' ]
                MaterialInfo[ 'ExtFlags' ] |= 1
            elif 'TextureCustom' in Material :
                MaterialInfo[ 'TextureID' ] = Material[ 'TextureCustom' ]
                MaterialInfo[ 'ExtFlags' ] |= 2
            
            if 'Collision' in Material :
                SurfaceIndex = GetSurfaceIndex( Material[ 'Collision' ] )
                
                if SurfaceIndex != -1 and SurfaceIndex != 27 :
                    MaterialInfo[ 'SurfaceID' ] = SurfaceIndex
                    MaterialInfo[ 'ExtFlags' ] |= 4

            if Object.data.show_double_sided :
                MaterialInfo[ 'ExtFlags' ] |= 8
            
            if 'NoTrails' in Material :
                MaterialInfo[ 'ExtFlags' ] |= 16
            
            if 'NoShadows' in Material :
                MaterialInfo[ 'ExtFlags' ] |= 32
            
            if 'IsTransparent' in Material :
                MaterialInfo[ 'ExtFlags' ] |= 64
            
            if MaterialInfo[ 'ExtFlags' ] == 0 :
                continue

            MaterialInfo[ 'ExtName' ] = 'm' + str( Index )
            MaterialInfo[ 'IntName' ] = Material.name
            ObjectMaterials.append( MaterialInfo )
        
        MaterialInfos.append( ObjectMaterials )
        MaterialsNb += len( ObjectMaterials )
    #
    
    Scene.objects.active = None
    Content = bytearray()
    Header = bytearray()    
    Instances = 1

    # MainNode
    WriteNumber( Content, 0x3F004000, 4 )
    
    WriteFloat( Content, SpawnX )
    WriteFloat( Content, SpawnY )
    WriteFloat( Content, SpawnZ )
    
    WriteStrLen( Content, BlockIdentifier )
    WriteStrLen( Content, BlockAuthor )
    
    WriteNumber( Content, MaterialsNb, 4 )
    
    for ObjectMaterials in MaterialInfos :
        
        for Material in ObjectMaterials :
            WriteNumber( Content, Material[ 'ExtFlags' ], 1 )
            WriteStrLen( Content, Material[ 'ExtName' ] )
            
            if Material[ 'ExtFlags' ] & 3 :
                WriteStrLen( Content, Material[ 'TextureID' ] )
            
            if Material[ 'ExtFlags' ] & 4 :
                WriteNumber( Content, Material[ 'SurfaceID' ], 1 )
    
    # Writing x000 chunk
    WriteNumber( Content, 0x09005000, 4 )
    WriteNumber( Content, 0x00000001, 4 )
    
    # Writing x011 chunk
    WriteNumber( Content, 0x09005011, 4 )
    WriteNumber( Content, 0x00000000, 4 )
    WriteNumber( Content, 0x00000000, 4 )    
    WriteNumber( Content, 0x00000001, 4 )
    ObjectIndex = -1
    
    for Index in range( 0, len( Objects ) ) :
        Object = Objects[ Index ]
        SeparatedObjects = []
    
        Object[ 'Identifier' ] = Object.name
        Scene.objects.active = Object
        
        if len( Object.data.materials ) > 1 :
            bpy.ops.object.mode_set( mode = 'EDIT' )
            bpy.ops.mesh.separate( type = 'MATERIAL' )
            bpy.ops.object.mode_set( mode = 'OBJECT' )
            
            for FinalObject in Context.visible_objects :
                if not 'Identifier' in FinalObject :
                    continue

                ObjectIndex += 1
                FinalObject[ 'Index' ] = ObjectIndex
                SeparatedObjects.append( FinalObject )

            Instances = DoWriteObject( Content, Object, SeparatedObjects, MaterialInfos[ Index ], Instances )
        else :
            Object[ 'Index' ] = 0
            Instances = DoWriteObject( Content, Object, [ Object ], MaterialInfos[ Index ], Instances )
    #

    bpy.ops.ed.undo()
    
    # Writing Header
    WriteStrRaw( Header, 'GBX' )
    WriteNumber( Header, 6, 2 )
    WriteStrRaw( Header, 'BUUR' )
    
    WriteNumber( Header, 0x3F004000, 4 )    
    WriteNumber( Header, 0x00000000, 4 )
    
    WriteNumber( Header, Instances + 2, 4 )
    WriteNumber( Header, 0x00000000, 4 )
    WriteNumber( Content, 0xFACADE01, 4 )    

    File = open( FilePath, 'wb' )
    File.write( Header )
    File.write( Content )
    File.close()
    
    return { 'FINISHED' }

from bpy.props import StringProperty, FloatProperty
from bpy_extras.io_utils import ExportHelper
from bpy.types import Operator

bl_info = {
    'name' : 'TrackMania Infinity Block Exporter',
    'author' : 'Tomek0055',
    'version' : ( 1, 2, 0 ),
    'blender' : ( 2, 78, 0 ),
    'location' : 'File > Export',
    'category' : 'Import-Export'
}

if 'bpy' in locals() :
    import importlib

    if 'export_gbx' in locals() :
        importlib.reload( export_gbx )

class GbxExporter ( Operator, ExportHelper ) :
    '''Save scene as TrackMania Infinity custom block'''
    bl_idname = 'export_scene.gbx'
    bl_label = 'Export'

    # ExportHelper mixin class uses this
    filename_ext = '.Block.Gbx'

    filter_glob = StringProperty (
        default = '*.Block.Gbx',
        options = { 'HIDDEN' },
    )

    BlockID = StringProperty (
        default = 'SampleBlock',
        name = 'Block ID',
        description = 'Block ID is used by game for identify your block in the world or while loading the map',
    )
    
    BlockAuthor = StringProperty (
        default = 'Nadeo',
        name = 'Author',
    )
    
    SpawnX = FloatProperty (
        name = 'Spawn X',
        default = 0.0,
    )
    
    SpawnY = FloatProperty (
        name = 'Spawn Y',
        default = 0.0,
    )
    
    SpawnZ = FloatProperty (
        name = 'Spawn Z',
        default = 0.0,
    )

    def execute ( Self, Context ) :
        return DoWriteBlock( Context, Self.filepath, Self.SpawnX, Self.SpawnY, Self.SpawnZ, Self.BlockID, Self.BlockAuthor )

def menu_func_export( Self, Context ) :
    Self.layout.operator( GbxExporter.bl_idname, text = 'TM Custom Block (.Block.Gbx)' )

def register():
    bpy.utils.register_class( GbxExporter )
    bpy.types.INFO_MT_file_export.append( menu_func_export )

def unregister():
    bpy.utils.unregister_class( GbxExporter )
    bpy.types.INFO_MT_file_export.remove( menu_func_export )

if __name__ == '__main__':
    register()
